/*
 * process, Distribution repositories and basic setup for Emmanuel Bigeon projects
 * Copyright (C) 2014-2018 E. Bigeon
 * mailto:emmanuel@bigeon.fr
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 *
 */
package net.bigeon.gclc.process.io;

/*-
 * #%L
 * process
 * %%
 * Copyright (C) 2014 - 2018 Bigeon
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.bigeon.gclc.manager.ConsoleInput;
import net.bigeon.gclc.tools.ConstantString;
import net.bigeon.gclc.tools.StringProvider;

/** A console input that can be connected to and diconnected from.
 * 
 * @author Emmanuel Bigeon */
public final class ConnectingConsoleInput implements ConsoleInput {

    /** The empty string provider. */
    private static final ConstantString EMPTY_STRING = new ConstantString("");
    /** The logger. */
    private static final Logger         LOGGER       = Logger
            .getLogger(ConnectingConsoleInput.class.getName());

    /** If the input is closed. */
    private boolean        close          = false;
    /** The prompt string. */
    private StringProvider prompt         = EMPTY_STRING;
    /** If the input is currently in prompting state.
     * <p>
     * To change it you should be in a promptLock. */
    private boolean        prompting      = false;
    /** The synchronization lock for the prompting status. */
    private final Object   promptLock     = new Object();
    /** The synchronization lock for the connection status. */
    private final Object   connectionLock = new Object();
    /** The connected console input.
     * <p>
     * To use it, you should be in a promptLock and connectionLock. */
    private ConsoleInput   connected      = null;
    /** The connection state.
     * <p>
     * To read or modify it, you should be in a connectionLock synchronize block. */
    private boolean        disconnection  = false;

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#close() */
    @Override
    public void close() {
        close = true;
    }

    /** Connect an input.
     * 
     * @param input the input to connect */
    public void connect(final ConsoleInput input) {
        disconnect();
        synchronized (promptLock) {
            connected = input;
            promptLock.notifyAll();
        }
    }

    /** Disconnect the current input. */
    public void disconnect() {
        synchronized (promptLock) {
            synchronized (connectionLock) {
                if (connected != null) {
                    disconnection = true;
                    connected.interruptPrompt();
                    connected = null;
                }
            }
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#getPrompt() */
    @Override
    public StringProvider getPrompt() {
        return prompt;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#interruptPrompt() */
    @Override
    public void interruptPrompt() {
        synchronized (promptLock) {
            prompting = false;
            if (connected != null) {
                connected.interruptPrompt();
            }
            promptLock.notifyAll();
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#isClosed() */
    @Override
    public boolean isClosed() {
        return close;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt() */
    @Override
    public String prompt() throws IOException {
        return prompt(prompt.apply());
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(long) */
    @Override
    public String prompt(final long timeout) throws IOException {
        return prompt(prompt.apply(), timeout);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(java.lang.String) */
    @Override
    public String prompt(final String message) throws IOException {
        synchronized (promptLock) {
            prompting = true;
        }
        while (true) {
            synchronized (promptLock) {
                if (!prompting) {
                    return null;
                }
                boolean connect;
                synchronized (connectionLock) {
                    connect = connected != null;
                }
                if (!connect) {
                    try {
                        promptLock.wait();
                    } catch (final InterruptedException e) {
                        LOGGER.log(Level.WARNING, "Inerruption of console thread", e);
                        Thread.currentThread().interrupt();
                    }
                } else {
                    synchronized (connectionLock) {
                        final String res = connected.prompt(message);
                        if (disconnection) {
                            disconnection = false;
                        } else if (prompting) {
                            return res;
                        } else {
                            // prompt interrupted, lose the result.
                        }
                    }
                }
            }
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(java.lang.String, long) */
    @Override
    public String prompt(final String message, final long timeout) throws IOException {
        if (timeout <= 0) {
            return prompt(message);
        }
        final long end = System.currentTimeMillis() + timeout;
        synchronized (promptLock) {
            prompting = true;
        }
        do {
            synchronized (promptLock) {
                if (!prompting) {
                    return null;
                }
                boolean connect;
                synchronized (connectionLock) {
                    connect = connected != null;
                }
                if (!connect) {
                    try {
                        promptLock.wait(timeout);
                    } catch (final InterruptedException e) {
                        LOGGER.log(Level.WARNING, "Inerruption of console thread", e);
                        Thread.currentThread().interrupt();
                    }
                } else {
                    synchronized (connectionLock) {
                        final String res = connected.prompt(message,
                                end - System.currentTimeMillis());
                        if (disconnection) {
                            disconnection = false;
                        } else if (prompting) {
                            return res;
                        } else {
                            // prompt interrupted, lose the result.
                        }
                    }
                }
            }
        } while (System.currentTimeMillis() < end);
        return null;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#setPrompt(java.lang.String) */
    @Override
    public void setPrompt(final String prompt) {
        this.prompt = new ConstantString(prompt);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#setPrompt(fr.bigeon.gclc.tools.
     * StringProvider) */
    @Override
    public void setPrompt(final StringProvider string) {
        prompt = string;
    }
}
